// Copyright 2020 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#ifndef MINI_CHROMIUM_BASE_NOTREACHED_H_
#define MINI_CHROMIUM_BASE_NOTREACHED_H_

#include "base/check.h"

#define NOTREACHED() DCHECK(false)

#endif  // MINI_CHROMIUM_BASE_NOTREACHED_H_
