#ifndef __HTM_HEADER_CODES_H__
#define __HTM_HEADER_CODES_H__

const char INSERT_KEYS = '1';
const char INIT_STATE = '2';
const char CLIENT_CLOSE_PANE = '3';
const char APPEND_TO_PANE = '4';
const char NEW_TAB = '5';
const char SERVER_CLOSE_PANE = '8';
const char NEW_SPLIT = '9';
const char RESIZE_PANE = 'A';
const char DEBUG_LOG = 'B';
const char INSERT_DEBUG_KEYS = 'C';
const char SESSION_END = 'D';

#endif