EXPORTS
    exif_array_set_byte_order
    exif_byte_order_get_name
    exif_content_add_entry
    exif_content_dump
    exif_content_fix
    exif_content_foreach_entry
    exif_content_free
    exif_content_get_entry
    exif_content_get_ifd
    exif_content_log
    exif_content_new
    exif_content_new_mem
    exif_content_ref
    exif_content_remove_entry
    exif_content_unref
    exif_data_dump
    exif_data_fix
    exif_data_foreach_content
    exif_data_free
    exif_data_get_byte_order
    exif_data_get_data_type
    exif_data_get_log
    exif_data_get_mnote_data
    exif_data_load_data
    exif_data_log
    exif_data_new
    exif_data_new_from_data
    exif_data_new_from_file
    exif_data_new_mem
    exif_data_option_get_description
    exif_data_option_get_name
    exif_data_ref
    exif_data_save_data
    exif_data_set_byte_order
    exif_data_set_data_type
    exif_data_set_option
    exif_data_unref
    exif_data_unset_option
    exif_entry_dump
    exif_entry_fix
    exif_entry_free
    exif_entry_get_value
    exif_entry_initialize
    exif_entry_new
    exif_entry_new_mem
    exif_entry_ref
    exif_entry_unref
    exif_format_get_name
    exif_format_get_size
    exif_get_long
    exif_get_rational
    exif_get_short
    exif_get_slong
    exif_get_srational
    exif_get_sshort
    exif_ifd_get_name
    exif_loader_get_buf
    exif_loader_get_data
    exif_loader_log
    exif_loader_new
    exif_loader_new_mem
    exif_loader_ref
    exif_loader_reset
    exif_loader_unref
    exif_loader_write
    exif_loader_write_file
    exif_log
    exif_log_code_get_message
    exif_log_code_get_title
    exif_log_free
    exif_log_new
    exif_log_new_mem
    exif_log_ref
    exif_log_set_func
    exif_log_unref
    exif_logv
    exif_mem_alloc
    exif_mem_free
    exif_mem_new
    exif_mem_new_default
    exif_mem_realloc
    exif_mem_ref
    exif_mem_unref
    exif_mnote_data_canon_new
    exif_mnote_data_construct
    exif_mnote_data_count
    exif_mnote_data_get_description
    exif_mnote_data_get_id
    exif_mnote_data_get_name
    exif_mnote_data_get_title
    exif_mnote_data_get_value
    exif_mnote_data_load
    exif_mnote_data_log
    exif_mnote_data_olympus_new
    exif_mnote_data_pentax_new
    exif_mnote_data_ref
    exif_mnote_data_save
    exif_mnote_data_set_byte_order
    exif_mnote_data_set_offset
    exif_mnote_data_unref
    exif_set_long
    exif_set_rational
    exif_set_short
    exif_set_slong
    exif_set_srational
    exif_set_sshort
    exif_tag_from_name
    exif_tag_get_description
    exif_tag_get_description_in_ifd
    exif_tag_get_name
    exif_tag_get_name_in_ifd
    exif_tag_get_support_level_in_ifd
    exif_tag_get_title
    exif_tag_get_title_in_ifd
    exif_tag_table_count
    exif_tag_table_get_name
    exif_tag_table_get_tag
    mnote_canon_entry_get_value
    mnote_canon_tag_get_description
    mnote_canon_tag_get_name
    mnote_canon_tag_get_title
    mnote_olympus_entry_get_value
    mnote_olympus_tag_get_description
    mnote_olympus_tag_get_name
    mnote_olympus_tag_get_title
    mnote_pentax_entry_get_value
    mnote_pentax_tag_get_description
    mnote_pentax_tag_get_name
    mnote_pentax_tag_get_title