XX(background_worker)
XX(basic_consent_tracking)
XX(basic_function_transport)
XX(basic_http_request_preparation_for_event)
XX(basic_http_request_preparation_for_event_with_attachment)
XX(basic_http_request_preparation_for_minidump)
XX(buildid_fallback)
XX(concurrent_init)
XX(count_sampled_events)
XX(custom_logger)
XX(dsn_parsing_complete)
XX(dsn_parsing_invalid)
XX(dsn_store_url_with_path)
XX(dsn_store_url_without_path)
XX(empty_transport)
XX(fuzz_json)
XX(init_failure)
XX(internal_uuid_api)
XX(invalid_dsn)
XX(invalid_proxy)
XX(iso_time)
XX(lazy_attachments)
XX(module_addr)
XX(module_finder)
XX(mpack_newlines)
XX(mpack_removed_tags)
XX(multiple_inits)
XX(os)
XX(page_allocator)
XX(path_basics)
XX(path_current_exe)
XX(path_directory)
XX(path_joining_unix)
XX(path_joining_windows)
XX(path_relative_filename)
XX(procmaps_parser)
XX(rate_limit_parsing)
XX(recursive_paths)
XX(sampling_before_send)
XX(serialize_envelope)
XX(session_basics)
XX(slice)
XX(symbolizer)
XX(task_queue)
XX(uninitialized)
XX(unwinder)
XX(url_parsing_complete)
XX(url_parsing_invalid)
XX(url_parsing_partial)
XX(uuid_api)
XX(uuid_v4)
XX(value_bool)
XX(value_collections_leak)
XX(value_double)
XX(value_freezing)
XX(value_int32)
XX(value_json_deeply_nested)
XX(value_json_escaping)
XX(value_json_invalid_doubles)
XX(value_json_locales)
XX(value_json_parsing)
XX(value_json_surrogates)
XX(value_list)
XX(value_null)
XX(value_object)
XX(value_string)
XX(value_unicode)
XX(value_wrong_type)
